﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace VitesseClientDemo.Json
{
    /// <summary>
    /// Process received JSON messages
    /// </summary>
    class JsonMessages
    {
        /// <summary>
        /// Callback to write formatted output
        /// </summary>
        public Action<string> Output { get; set; }

        /// <summary>
        /// Callback to write error output
        /// </summary>
        public Action<string> Error { get; set; }

        #region Keys
        private const string key_ageclass = "ageclass";
        private const string key_competition = "competition";
        private const string key_country = "country";
        private const string key_crossing = "crossing";
        private const string key_date = "date";
        private const string key_distance = "distance";
        private const string key_end = "end";
        private const string key_familyname = "familyname";
        private const string key_finishing = "finishing";
        private const string key_gender = "gender";
        private const string key_givenname = "givenname";
        private const string key_helmet = "helmet";
        private const string key_id = "id";
        private const string key_info = "info";
        private const string key_inner = "inner";
        private const string key_location = "location";
        private const string key_male = "male";
        private const string key_msgoptions = "msgoptions";
        private const string key_name = "name";
        private const string key_new = "new";
        private const string key_number = "number";
        private const string key_old = "old";
        private const string key_outer = "outer";
        private const string key_quad = "quad";
        private const string key_racedel = "racedel";
        private const string key_races = "races";
        private const string key_skater = "skater";
        private const string key_skaterdel = "skaterdel";
        private const string key_skaternum = "skaternum";
        private const string key_skaters = "skaters";
        private const string key_start = "start";
        private const string key_startlist = "startlist";
        private const string key_suffix = "suffix";
        private const string key_track = "track";
        private const string key_type = "type";
        private const string key_venue = "venue";
        #endregion

        #region Values
        private const string val_pairs = "pairs";
        private const string val_massstart = "massstart";
        private const string val_teampursuit = "teampursuit";
        private const string val_teamsprint = "teamsprint";
        #endregion

        private string separator = "-------------------";

        /// <summary>
        /// Initialize a JSON message handler
        /// </summary>
        /// <param name="output"></param>
        /// <param name="error"></param>
        public JsonMessages(Action<string> output, Action<string> error = null)
        {
            Output = output;
            Error = error;
        }

        #region Receive
        private void OutputLine(string line)
        {
            Output(line + "\n");
        }

        private void OutputHeader(string header)
        {
            OutputLine("========== " + header + " ==========");
        }

        private void OutputSubheader(string header)
        {
            OutputLine("---------- " + header + " ----------");
        }

        /// <summary>
        /// Parse a message
        /// </summary>
        /// <param name="msg"></param>
        public void Parse(JObject msg)
        {
            try
            {
                if (msg.ContainsKey(key_competition)) { 
                    OutputHeader("Competition");
                    CompetitionParse(JsonParser.GetObject(msg, key_competition));
                } else if (msg.ContainsKey(key_info)) { 
                    OutputHeader("Info");
                    InfoParse(JsonParser.GetObject(msg, key_info));
                } else if (msg.ContainsKey(key_races)) { 
                    OutputHeader("Races");
                    RacesParse(JsonParser.GetArray(msg, key_races));
                } else if (msg.ContainsKey(key_racedel))
                {
                    OutputHeader("Race Delete");
                    RaceDelParse(JsonParser.GetArray(msg, key_racedel));
                } else if (msg.ContainsKey(key_skaters))
                {
                    OutputHeader("Skaters");
                    ParticipantsParse(JsonParser.GetArray(msg, key_skaters));
                } else if (msg.ContainsKey(key_skaternum))
                {
                    OutputHeader("Renumber Skaters");
                    RenumberParse(JsonParser.GetArray(msg, key_skaternum));
                } else if (msg.ContainsKey(key_skaterdel))
                {
                    OutputHeader("Skater Delete");
                    SkaterDelParse(JsonParser.GetArray(msg, key_skaterdel));
                }
            } catch (Exception e)
            {
                if (Error != null)
                {
                    Error("JSON error: " + e.Message);
                }
            }
        }

        /// <summary>
        /// Parse a competition message
        /// </summary>
        /// <param name="competition"></param>
        private void CompetitionParse(JObject competition)
        {
            OutputSubheader("Info");
            JObject info = JsonParser.GetObject(competition, key_info, "competition.info");
            InfoParse(info);

            JArray skaters = JsonParser.GetArray(competition, key_skaters, required: false);
            if (skaters != null)
            {
                OutputSubheader("Skaters");
                ParticipantsParse(skaters);
            }

            // Races
            JArray races = JsonParser.GetArray(competition, key_races, required: false);
            if (races != null)
            {
                OutputSubheader("Races");
                RacesParse(races);
            }
        }

        /// <summary>
        /// Parse an info message
        /// </summary>
        /// <param name="info"></param>
        private void InfoParse(JObject info)
        {
            string name = JsonParser.GetString(info, key_name, "info.name");
            DateTime startDate = JsonParser.GetDate(info, key_start, "info.startdate").Value;
            DateTime endDate = JsonParser.GetDate(info, key_end, "info.enddate").Value;

            JObject track = JsonParser.GetObject(info, key_track, "info.track");
            string location = JsonParser.GetString(track, key_location, "track.location");
            string venue = JsonParser.GetString(track, key_venue, fallback: "");

            OutputLine(name);
            OutputLine(DateFormatter.Dates(startDate, endDate));
            OutputLine(location);
            if (!string.IsNullOrWhiteSpace(venue))
            {
                OutputLine(venue);
            }
        }

        /// <summary>
        /// Parse a list of skaters
        /// </summary>
        /// <param name="skaters"></param>
        private void ParticipantsParse(JArray skaters)
        {
            OutputLine("Count = " + skaters.Count);
            foreach (var skater in skaters)
            {
                OutputLine(ParticipantParse(skater));
            }
        }

        /// <summary>
        /// Parse and format a Participant
        /// </summary>
        /// <param name="participant"></param>
        /// <returns>formatted info</returns>
        private string ParticipantParse(JToken participant)
        {
            string number = JsonParser.GetUInt(participant, key_number, "skater.number").ToString().PadRight(5);

            string name = (JsonParser.GetString(participant, key_familyname, fallback: "").ToUpper() + " " +
                JsonParser.GetString(participant, key_givenname, fallback: "")).Trim();
            string suffix = JsonParser.GetString(participant, key_suffix, fallback: "");
            if (!string.IsNullOrWhiteSpace(suffix))
            {
                name += " (" + suffix + ")";
            }
            name = name.PadRight(60).Substring(0, 60);

            string country = JsonParser.GetString(participant, key_country, "skater.country");

            string ageclass = JsonParser.GetString(participant, key_ageclass, fallback: "");
            bool male = JsonParser.GetBool(participant, key_male, "skater.male");
            ageclass = ((male ? "M" : "L") + ageclass).PadRight(3);

            return number + " " + name + " " + country + " " + ageclass;
        }

        /// <summary>
        /// Parse a list of races
        /// </summary>
        /// <param name="races"></param>
        private void RacesParse(JArray races)
        {
            OutputLine("Count = " + races.Count);
            foreach (var race in races)
            {
                RaceParse(race);
            }
        }

        /// <summary>
        /// Parse a single race
        /// </summary>
        /// <param name="race"></param>
        private void RaceParse(JToken race)
        {
            string type = JsonParser.GetString(race, key_type, "race.type");
            int id = JsonParser.GetInt(race, key_id, "race.id");
            int distance = JsonParser.GetInt(race, key_distance, "race.distance");
            string gender = JsonParser.GetString(race, key_gender, "race.gender");
            DateTime date = JsonParser.GetDate(race, key_date, "race.date").Value;
            string name = JsonParser.GetString(race, key_name, "race.name");
            JArray startlist = JsonParser.GetArray(race, key_startlist, "race.startlist");

            OutputLine(separator);
            OutputLine(id + ": " + name + " - " + DateFormatter.Date(date));

            if (type.Equals(val_pairs))
            {
                PairsParse(startlist);
            } else if (type.Equals(val_massstart))
            {
                MassStartParse(startlist);
            } else if (type.Equals(val_teampursuit) || type.Equals(val_teamsprint))
            {
                TeamPairsParse(startlist);
            }
            else
            {
                if (Error != null)
                {
                    Error("Unknown race.type " + type);
                }
            }
            // start list
        }

        /// <summary>
        /// Parse deleted races
        /// </summary>
        /// <param name="raceIds"></param>
        private void RaceDelParse(JArray raceIds)
        {
            OutputLine("Count = " + raceIds.Count);
            Output("Race IDs: ");
            string ids = "";
            foreach (var id in raceIds)
            {
                ids += (!string.IsNullOrEmpty(ids) ? ", " : "") + id;
            }
            OutputLine(ids);
        }

        /// <summary>
        /// Parse a pairs list
        /// </summary>
        /// <param name="pairs"></param>
        private void PairsParse(JArray pairs)
        {
            OutputLine(pairs.Count + " Pairs");
            foreach (var pair in pairs)
            {
                PairParse(pair);
            }
        }

        /// <summary>
        /// Parse a single pair
        /// </summary>
        /// <param name="pair"></param>
        private void PairParse(JToken pair)
        {
            string number = JsonParser.GetInt(pair, key_number, "pair.number").ToString().PadRight(3);
            int quad = JsonParser.GetInt(pair, key_quad, fallback: 0);
            JObject inner = JsonParser.GetObject(pair, key_inner, required: false);
            JObject outer = JsonParser.GetObject(pair, key_outer, required: false);

            string skaterInner = (inner != null) ? ParticipantParse(inner) : "";
            string skaterOuter = (outer != null) ? ParticipantParse(outer) : "";

            string laneInner, laneOuter;
            switch (quad)
            {
                case 1:
                    laneInner = "W";
                    laneOuter = "R";
                    break;

                case 2:
                    laneInner = "Y";
                    laneOuter = "B";
                    break;

                default:
                    laneInner = "I";
                    laneOuter = "O";
                    break;
            }


            OutputLine(number + " " + laneInner + " " + skaterInner);
            OutputLine("    " + laneOuter + " " + skaterOuter);
            if (quad != 1)
            {
                OutputLine("");
            }
        }

        /// <summary>
        /// Parse a mass start start list
        /// </summary>
        /// <param name="startlist"></param>
        private void MassStartParse(JArray startlist)
        {
            string number;
            string helmet;
            JObject skaterJson;
            string skater;

            OutputLine(startlist.Count + " Skaters");
            foreach (var position in startlist)
            {
                number = JsonParser.GetInt(position, key_number).ToString().PadRight(3);
                helmet = JsonParser.GetInt(position, key_helmet).ToString().PadRight(3);
                skaterJson = JsonParser.GetObject(position, key_skater, required: false);

                skater = (skaterJson != null) ? ParticipantParse(JsonParser.GetObject(position, key_skater)) : "";

                OutputLine(number + " " + helmet + " " + skater);
            }
        }

        /// <summary>
        /// Parse a pairs list
        /// </summary>
        /// <param name="pairs"></param>
        private void TeamPairsParse(JArray pairs)
        {
            OutputLine(pairs.Count + " Pairs");
            foreach (var pair in pairs)
            {
                TeamPairParse(pair);
            }
        }

        /// <summary>
        /// Parse a single team pair
        /// </summary>
        /// <param name="pair"></param>
        private void TeamPairParse(JToken pair)
        {
            string number = JsonParser.GetInt(pair, key_number, "pair.number").ToString().PadRight(3);
            JObject finishing = JsonParser.GetObject(pair, key_finishing, required: false);
            JObject crossing = JsonParser.GetObject(pair, key_crossing, required: false);

            string teamFinishing = (finishing != null) ? TeamParse(finishing) : "";
            string teamCrossing = (crossing != null) ? TeamParse(crossing) : "";

            string laneFinishing = "F", laneCrossing = "C";

            OutputLine(number + " " + laneFinishing + " " + teamFinishing);
            OutputLine("    " + laneCrossing + " " + teamCrossing);
            OutputLine("");
        }

        private string TeamParse(JObject team)
        {
            string number = JsonParser.GetInt(team, key_number, "team.number").ToString().PadRight(5);
            string name = JsonParser.GetString(team, key_name, "team.name", "");
            string text = number + " " + name;

            JArray skaters = JsonParser.GetArray(team, key_skaters, required: false);
            if (skaters != null)
            {
                foreach (var skater in skaters)
                {
                    text += "\n            " + ParticipantParse(skater);
                }
            }

            return text;
        }

        /// <summary>
        /// Parse a skater renumber message
        /// </summary>
        /// <param name="changes"></param>
        private void RenumberParse(JArray changes)
        {
            uint oldNum;
            uint newNum;

            OutputLine("Count = " + changes.Count);
            foreach (var change in changes)
            {
                oldNum = JsonParser.GetUInt(change, key_old);
                newNum = JsonParser.GetUInt(change, key_new);
                OutputLine(oldNum + " => " + newNum);
            }
        }

        /// <summary>
        /// Parse deleted skaters
        /// </summary>
        /// <param name="skaterIds"></param>
        private void SkaterDelParse(JArray skaterIds)
        {
            OutputLine("Count = " + skaterIds.Count);
            Output("Skater IDs: ");
            string ids = "";
            foreach (var id in skaterIds)
            {
                ids += (!string.IsNullOrEmpty(ids) ? ", " : "") + id;
            }
            OutputLine(ids);
        }
        #endregion

        #region Send
        /// <summary>
        /// Build a message options message
        /// </summary>
        /// <param name="info"></param>
        /// <param name="skaters"></param>
        /// <param name="races"></param>
        /// <returns>message</returns>
        public string MsgOptionsBuild(bool info = true, bool skaters = true, bool races = true)
        {
            return new JObject(new JProperty(key_msgoptions, new JObject(
                new JProperty(key_info, info),
                new JProperty(key_skaters, skaters),
                new JProperty(key_races, races)))).ToString(Formatting.None);
        }
        #endregion
    }
}
