﻿using System;
using Newtonsoft.Json.Linq;

namespace VitesseClientDemo.Json
{
    class JsonParser
    {
        /// <summary>
        /// Get a JSON object
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="required">presence required?</param>
        /// <returns>JObject</returns>
        static public JObject GetObject(JToken json, string key, string errInfo = null, bool required = true)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            if (json[key] != null)
            {
                JObject obj = json[key] as JObject;
                if (obj != null)
                {
                    return obj;
                }
                else
                {
                    throw new ParseError(errInfo + " not valid");
                }
            }
            else
            {
                if (required)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    return null;
                }
            }
        }

        /// <summary>
        /// Get a JSON array
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="required">presence required?</param>
        /// <returns>JObject</returns>
        static public JArray GetArray(JToken json, string key, string errInfo = null, bool required = true)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            if (json[key] != null)
            {
                JArray array = json[key] as JArray;
                if (array != null)
                {
                    return array;
                }
                else
                {
                    throw new ParseError(errInfo + " not valid");
                }
            }
            else
            {
                if (required)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    return null;
                }
            }
        }

        /// <summary>
        /// Get a string value
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="fallback">fallback value, null == none</param>
        /// <returns></returns>
        static public string GetString(JToken json, string key,
            string errInfo = null, string fallback = null)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            string value;

            if (json[key] != null)
            {
                value = (string)json[key];
            }
            else
            {
                if (fallback == null)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    value = fallback;
                }
            }

            return value;
        }

        /// <summary>
        /// Get a ushort value
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="fallback">fallback value, null == none</param>
        /// <returns></returns>
        static public ushort GetUShort(JToken json, string key, string errInfo = null, ushort? fallback = null)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            ushort value;

            if (json[key] != null)
            {
                value = (ushort)json[key];
            }
            else
            {
                if (fallback == null)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    value = fallback.Value;
                }
            }

            return value;
        }

        /// <summary>
        /// Get a uint value
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="fallback">fallback value, null == none</param>
        /// <returns></returns>
        static public uint GetUInt(JToken json, string key, string errInfo = null, uint? fallback = null)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            uint value;

            if (json[key] != null)
            {
                value = (uint)json[key];
            }
            else
            {
                if (fallback == null)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    value = fallback.Value;
                }
            }

            return value;
        }

        /// <summary>
        /// Get an int value
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="fallback">fallback value, null == none</param>
        /// <returns></returns>
        static public int GetInt(JToken json, string key, string errInfo = null, int? fallback = null)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            int value;

            if (json[key] != null)
            {
                value = (int)json[key];
            }
            else
            {
                if (fallback == null)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    value = fallback.Value;
                }
            }

            return value;
        }

        /// <summary>
        /// Get a bool value
        /// </summary>
        /// <param name="json">JSON</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="fallback">fallback value, null == none</param>
        /// <returns></returns>
        static public bool GetBool(JToken json, string key, string errInfo = null, bool? fallback = null)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            bool value;

            if (json[key] != null)
            {
                value = (bool)json[key];
            }
            else
            {
                if (fallback == null)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    value = fallback.Value;
                }
            }

            return value;
        }

        /// <summary>
        /// Get a date
        /// </summary>
        /// <param name="json">JSON object</param>
        /// <param name="key">key</param>
        /// <param name="errInfo">error message info</param>
        /// <param name="required">presence required?</param>
        /// <returns>Date</returns>
        static public DateTime? GetDate(JToken json, string key, string errInfo = null, bool required = true)
        {
            if (errInfo == null)
            {
                errInfo = key;
            }

            if (json[key] != null)
            {
                try
                {
                    return DateTime.Parse((string)json[key]);
                }
                catch
                {
                    throw new ParseError(errInfo + " not valid");
                }
            }
            else
            {
                if (required)
                {
                    throw new ParseError(errInfo + " not present");
                }
                else
                {
                    return null;
                }
            }
        }

        public class ParseError : Exception
        {
            public ParseError(string message) : base(message) { }
        }
    }
}
